/***************************************************************************
 *   Copyright (C) 2008 by Artur Trzewik   *
 *   mail@xdobry.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <cppuhelper/implbase2.hxx> // "3" implementing three interfaces
#include <cppuhelper/factory.hxx>
#include <rtl/ustring.hxx>
#include <osl/file.hxx>
#include <cppuhelper/queryinterface.hxx> // helper for queryInterface() impl

#include <com/sun/star/lang/XSingleServiceFactory.hpp>
#include <com/sun/star/lang/XMultiServiceFactory.hpp>
#include <com/sun/star/lang/XServiceInfo.hpp>
#include <com/sun/star/lang/XInitialization.hpp>
#include <com/sun/star/linguistic2/XProofreader.hpp>
#include <com/sun/star/lang/IllegalArgumentException.hpp>
#include <com/sun/star/lang/Locale.hpp>
#include <com/sun/star/registry/XRegistryKey.hpp>
#include <com/sun/star/deployment/XPackageInformationProvider.hpp>
#include <com/sun/star/deployment/PackageInformationProvider.hpp>
#include <com/sun/star/text/TextMarkupType.hdl>

#include <com/sun/star/awt/XToolkit.hdl>
#include <com/sun/star/awt/XMessageBoxFactory.hdl>
#include <com/sun/star/awt/XWindowPeer.hdl>
#include <com/sun/star/awt/XMessageBox.hdl>
#include <com/sun/star/awt/MessageBoxButtons.hdl>
#include <com/sun/star/frame/XDesktop.hdl>
#include <com/sun/star/frame/XFrame.hdl>

#include <string.h>
#include <tcl.h>

#define SERVICENAME "esperantilo.Korektilo"
#define PROOFSERVICE "com.sun.star.linguistic2.Proofreader"
#define IMPLNAME	"esperantilo.parto.Korektilo"
#define IDENTIFIER "org.openoffice.esperantilooo.oxt"


// ostring.pData->buffer

namespace css = ::com::sun::star;

using namespace ::rtl;
using namespace ::osl;
using namespace ::cppu;
using namespace ::rtl; // for OUString
using namespace ::com::sun::star; // for sdk interfaces
using namespace ::com::sun::star::uno; // for basic types
using namespace ::com::sun::star::lang; // for basic types
using namespace ::com::sun::star::linguistic2;
using namespace ::com::sun::star::registry;

EXTERN int Xotcl_Init _ANSI_ARGS_((Tcl_Interp * interp));

class MyGrammarChecker : public 
	::cppu::WeakImplHelper2<lang::XServiceInfo,XProofreader> {
		
	OUString m_sData;
    Reference< XComponentContext > m_xContext;
	Tcl_Interp *interp; 
	Tcl_Encoding encoding;
	Tcl_Obj *checkCommand;
	// File Logger
	File *myfile;
private:
	void initChecker();
    void initLogger(OUString parentDir);
	void writeLog(char *log);
	Tcl_Obj *oustringToTclString(const OUString &oustring);
	OUString tclStringToOUString(Tcl_Obj *obj);
public:
    inline MyGrammarChecker(Reference< XComponentContext > const & xContext) throw ()
        : m_xContext(xContext) {
		// printf("create esp instance\n");
		initChecker();
	}

    virtual ~MyGrammarChecker() {
		// printf("free esp instance\n");		
		Tcl_FreeEncoding(encoding);
		Tcl_DecrRefCount(checkCommand);
		Tcl_DeleteInterp(interp);
		if (myfile) myfile->close();
	}

    // ::com::sun::star::linguistic2::XSupportedLocales:
    virtual css::uno::Sequence< css::lang::Locale > SAL_CALL getLocales() throw (css::uno::RuntimeException);
    virtual ::sal_Bool SAL_CALL hasLocale(const css::lang::Locale & aLocale) throw (css::uno::RuntimeException);


    // ::com::sun::star::linguistic2::XGrammarChecker:

    virtual ::sal_Bool SAL_CALL isSpellChecker() throw (css::uno::RuntimeException);


    virtual ::com::sun::star::linguistic2::ProofreadingResult SAL_CALL doProofreading( const ::rtl::OUString& aDocumentIdentifier, const ::rtl::OUString& aText, const ::com::sun::star::lang::Locale& aLocale, ::sal_Int32 nStartOfSentencePosition, ::sal_Int32 nSuggestedBehindEndOfSentencePosition, const ::com::sun::star::uno::Sequence< ::com::sun::star::beans::PropertyValue >& aProperties ) throw (::com::sun::star::lang::IllegalArgumentException, ::com::sun::star::uno::RuntimeException);

    virtual void SAL_CALL ignoreRule( const ::rtl::OUString& aRuleIdentifier, const ::com::sun::star::lang::Locale& aLocale )         throw (::com::sun::star::lang::IllegalArgumentException, ::com::sun::star::uno::RuntimeException);

	virtual void SAL_CALL resetIgnoreRules(  ) throw (::com::sun::star::uno::RuntimeException);

    // XServiceInfo
    virtual OUString SAL_CALL getImplementationName() throw (RuntimeException);
    virtual sal_Bool SAL_CALL supportsService( OUString const & serviceName )
        throw (RuntimeException);
    virtual Sequence< OUString > SAL_CALL getSupportedServiceNames()
        throw (RuntimeException);

	static Sequence< OUString > SAL_CALL getSupportedServiceNames_Static(  );

};

void MyGrammarChecker::initLogger(OUString parentDir) {
	OUString url = parentDir+OUString::createFromAscii("/")+OUString::createFromAscii("log.txt");
	OString ostring = OUStringToOString( url, RTL_TEXTENCODING_ASCII_US );
	OSL_TRACE("init logger");
	OSL_TRACE(ostring.pData->buffer);
	myfile = new File(url);
	myfile->open(osl_File_OpenFlag_Write | osl_File_OpenFlag_Create);
	sal_uInt64 succes;
}

void MyGrammarChecker::writeLog(char *log) {
	sal_uInt64 succes;
	OSL_TRACE(log);
	myfile->write(log,strlen(log),succes);
	myfile->write("\n",1,succes);
}


void MyGrammarChecker::initChecker() {
	Tcl_Obj *path;

	// printf("init checker\n");		
	css::uno::Reference< css::deployment::XPackageInformationProvider >
  	xInfoProvider( css::deployment::PackageInformationProvider::get( m_xContext ) );
 
  	// "MY_PACKAGE_ID" is the identifier of the package whose location we want
  	OUString sLocation = xInfoProvider->getPackageLocation( 
  	OUString::createFromAscii( IDENTIFIER ) );
	OUString systemPath;
	FileBase::getSystemPathFromFileURL(sLocation,systemPath);
	OString ostring = OUStringToOString( systemPath, RTL_TEXTENCODING_ASCII_US );

	initLogger(sLocation);

    interp = Tcl_CreateInterp();

	// printf("progdir %s\n",ostring.getStr());
	
	Tcl_SetVar(interp, "progdir", ostring.getStr(), TCL_GLOBAL_ONLY);
	Xotcl_Init(interp);

	Tcl_Obj *lobj = Tcl_NewListObj(0, NULL);
	Tcl_Obj *systemPathObj = Tcl_NewStringObj(ostring.getStr(), ostring.getLength());
	
	Tcl_SetEncodingSearchPath(systemPathObj);
	encoding = Tcl_GetEncoding(interp, "unicode");

	Tcl_ListObjAppendElement(NULL, lobj, systemPathObj);
	Tcl_ListObjAppendElement(NULL, lobj,  Tcl_NewStringObj("main.tcl", -1));
	path = Tcl_FSJoinPath(lobj,-1);
	Tcl_IncrRefCount(path);
	// std::cout << "main.tcl path: " << Tcl_GetStringFromObj(path, 0) << "\n";
	if (Tcl_FSEvalFileEx(interp,path,"utf-8")==TCL_ERROR) {
		writeLog ("esperantilo tcl init error: '");
		writeLog (Tcl_GetStringFromObj(Tcl_GetObjResult(interp),0));
	}
	Tcl_DecrRefCount(path);
	checkCommand = Tcl_NewStringObj("doCheck", -1);
	Tcl_IncrRefCount(checkCommand);
	writeLog("Esperantilo initialized");
}

Tcl_Obj *MyGrammarChecker::oustringToTclString(const OUString &oustring) {
	Tcl_Obj *obj;
  	Tcl_DString ds;
    // We do not use  oustring.getLength() because tcl need byte length
	Tcl_ExternalToUtfDString(encoding, (const char*)oustring.getStr(), -1, &ds);
    obj = Tcl_NewStringObj(Tcl_DStringValue(&ds), Tcl_DStringLength(&ds));
    Tcl_DStringFree(&ds);
	return obj;
}
OUString MyGrammarChecker::tclStringToOUString(Tcl_Obj *obj) {
	Tcl_DString queryDS;
	char *query;
    int queryLen;

	query=Tcl_GetStringFromObj(obj, &queryLen);
	Tcl_UtfToExternalDString(encoding, query, -1, &queryDS);
    queryLen = Tcl_DStringLength(&queryDS);
	OUString ret((const sal_Unicode *)Tcl_DStringValue(&queryDS),Tcl_GetCharLength(obj));
    Tcl_DStringFree(&queryDS);
	return ret;
}


// XGrammarCheckerImplemenatation

void SAL_CALL MyGrammarChecker::ignoreRule( const ::rtl::OUString& aRuleIdentifier, const ::com::sun::star::lang::Locale& aLocale )         throw (::com::sun::star::lang::IllegalArgumentException, ::com::sun::star::uno::RuntimeException) {
    // printf("ignoreRule\n");
}

void SAL_CALL MyGrammarChecker::resetIgnoreRules(  ) throw (::com::sun::star::uno::RuntimeException) {
    // printf("resetIgnoreRules\n");
}

::sal_Bool SAL_CALL MyGrammarChecker::isSpellChecker() throw (RuntimeException) {
    // printf("isSpellChecker\n");
	return sal_False;
}

ProofreadingResult SAL_CALL MyGrammarChecker::doProofreading( const OUString& aDocumentIdentifier, const OUString& aText, const Locale& aLocale, ::sal_Int32 nStartOfSentencePosition, ::sal_Int32 nSuggestedBehindEndOfSentencePosition, const Sequence< ::com::sun::star::beans::PropertyValue >& aProperties ) throw (IllegalArgumentException, RuntimeException)
{
    // printf("doProofreading\n");
	ProofreadingResult paRes;
    paRes.xProofreader = this;
    paRes.aLocale = aLocale;
    paRes.aDocumentIdentifier = aDocumentIdentifier;
    paRes.aText = aText;
    paRes.nStartOfSentencePosition = nStartOfSentencePosition;
	paRes.aProperties = aProperties;

	Tcl_Obj *objv[4];
	// objv[0] = Tcl_NewStringObj("doCheck", -1);
	objv[0] = checkCommand;
	objv[1] = oustringToTclString(aText);
	objv[2] = Tcl_NewIntObj(nStartOfSentencePosition);
	objv[3] = Tcl_NewIntObj(nSuggestedBehindEndOfSentencePosition);
	for (int x=0;x<4;x++) {
		Tcl_IncrRefCount(objv[x]);
	}
    // printf("c docheck beginn\n");
	if (Tcl_EvalObjv(interp, 4, objv, TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT)==TCL_ERROR) {
	    // printf("c docheck end\n");
		writeLog ( "esperantilo doCheck error: '");
		writeLog (Tcl_GetStringFromObj(Tcl_GetObjResult(interp),0));
		writeLog("tcl stack trace:");
		Tcl_Obj *error = Tcl_GetVar2Ex(interp, "errorInfo", NULL, TCL_GLOBAL_ONLY);
		if (error) {
			writeLog(Tcl_GetStringFromObj(error,0));
		} else {
			writeLog("no error Info");
		}
		paRes.nBehindEndOfSentencePosition = nSuggestedBehindEndOfSentencePosition;
		paRes.nStartOfNextSentencePosition = nSuggestedBehindEndOfSentencePosition;
		for (int x=0;x<4;x++) {
			Tcl_DecrRefCount(objv[x]);
		}
		return paRes;
	}
    // printf("c docheck end\n");
	for (int x=0;x<4;x++) {
		Tcl_DecrRefCount(objv[x]);
	}
	
	Tcl_Obj **listObjv,**listErrorv,**errorv,**suggestionsv,*res;
	int listObjc,listErrorc,errorc,suggestionsc;
	res = Tcl_GetObjResult(interp);
	if (Tcl_ListObjGetElements(interp, res, &listObjc, &listObjv) != TCL_OK) {
		writeLog ("esperantilo can not convert doCheck result: '");
		writeLog (Tcl_GetStringFromObj(Tcl_GetObjResult(interp),0));
		paRes.nBehindEndOfSentencePosition = nSuggestedBehindEndOfSentencePosition;
		paRes.nStartOfNextSentencePosition = nSuggestedBehindEndOfSentencePosition;
		return paRes;
	}
 	if (listObjc!=3) {
		paRes.nBehindEndOfSentencePosition = nSuggestedBehindEndOfSentencePosition;
		paRes.nStartOfNextSentencePosition = nSuggestedBehindEndOfSentencePosition;
		return paRes;
	}
	// paRes.aText = tclStringToOUString(listObjv[0]);
	// OString ostring2 = OUStringToOString( paRes.aText, RTL_TEXTENCODING_ASCII_US );
	// std::cout << "esperantilo doCheck text: '" << ostring2 << "'\n";

	int endPos;
	Tcl_GetIntFromObj(interp, listObjv[1], &endPos);
	paRes.nBehindEndOfSentencePosition = endPos;
	paRes.nStartOfNextSentencePosition = endPos;
	if (Tcl_ListObjGetElements(interp, listObjv[2], &listErrorc, &listErrorv) != TCL_OK) {
		writeLog ("esperantilo can not convert doCheck result: '");
		writeLog (Tcl_GetStringFromObj(Tcl_GetObjResult(interp),0));
		return paRes;
	}
    Sequence< SingleProofreadingError > errors(listErrorc);
	for (int x=0;x<listErrorc;x++) {
		SingleProofreadingError ierror;
		if (Tcl_ListObjGetElements(interp, listErrorv[x], &errorc, &errorv) != TCL_OK || errorc!=6) {
			writeLog ("esperantilo can not convert doCheck result: '");
			writeLog (Tcl_GetStringFromObj(Tcl_GetObjResult(interp),0));
			return paRes;
		}
		int intv;
		Tcl_GetIntFromObj(interp, errorv[0], &intv);
		ierror.nErrorStart = intv;
		Tcl_GetIntFromObj(interp, errorv[1], &intv);
		ierror.nErrorLength = intv;
		Tcl_GetIntFromObj(interp, errorv[2], &intv);
		// ierror.nErrorLevel = intv;
		ierror.nErrorType = text::TextMarkupType::PROOFREADING;
		ierror.aShortComment = tclStringToOUString(errorv[3]);
		
		OString ostring2 = OUStringToOString( ierror.aShortComment, RTL_TEXTENCODING_ASCII_US );
		// std::cout << "esperantilo doCheck suggestion: '" << ostring2 << "'\n";

		ierror.aFullComment = tclStringToOUString(errorv[4]);
		if (Tcl_ListObjGetElements(interp, errorv[5], &suggestionsc, &suggestionsv)==TCL_OK) {
			Sequence<OUString> suggestions(suggestionsc);
			for(int i=0;i<suggestionsc;i++) {
				suggestions[i] = tclStringToOUString(suggestionsv[i]);
			}
			ierror.aSuggestions = suggestions;
		}
		errors[x] = ierror;
	}
	if (listErrorc>0) {
		paRes.aErrors = errors;
	}
	paRes.aText = aText;
	paRes.nBehindEndOfSentencePosition = nSuggestedBehindEndOfSentencePosition;
	paRes.nStartOfNextSentencePosition = nSuggestedBehindEndOfSentencePosition;
	return paRes;
}

// XSupportedLocales

Sequence< lang::Locale > SAL_CALL MyGrammarChecker::getLocales(  ) throw (RuntimeException) {
	// printf("getLocales\n");
	lang::Locale nLoc( OUString::createFromAscii("eo"),OUString(), OUString() );
	Sequence< lang::Locale > aLocales(&nLoc,1);
	return aLocales;
}

::sal_Bool SAL_CALL MyGrammarChecker::hasLocale( const lang::Locale& aLocale ) throw (RuntimeException) {
	// printf("hasLocale\n");
	return aLocale.Language.equalsAscii("eo");
}



// XServiceInfo implementation
OUString MyGrammarChecker::getImplementationName()
    throw (RuntimeException)
{
    // unique implementation name
    return OUString( RTL_CONSTASCII_USTRINGPARAM(IMPLNAME) );
}

sal_Bool MyGrammarChecker::supportsService( OUString const & serviceName )
    throw (RuntimeException)
{
    return serviceName.equalsAsciiL( RTL_CONSTASCII_STRINGPARAM(PROOFSERVICE)) ;
}

Sequence< OUString > MyGrammarChecker::getSupportedServiceNames()
    throw (RuntimeException)
{
    return getSupportedServiceNames_Static();
}

static uno::Reference<uno::XInterface> laKorektilo;

Reference< XInterface > SAL_CALL create_KorektiloImpl(
    Reference< XComponentContext > const & xContext )
    SAL_THROW( () )
{
	if (!laKorektilo.is()) {
    	// printf("create_KorektiloImpl\n");
        laKorektilo = static_cast< OWeakObject * >(new MyGrammarChecker(xContext));
    }
    return laKorektilo;
}

Sequence<OUString> SAL_CALL MyGrammarChecker::getSupportedServiceNames_Static(  ) 
{
	Sequence < OUString > aRet(1);
    OUString* pArray = aRet.getArray();
    pArray[0] =  OUString ( RTL_CONSTASCII_USTRINGPARAM ( PROOFSERVICE ) );
    return aRet;
}	


//#########################################################################
//#### EXPORTED ###########################################################
//#########################################################################


/**
 * Gives the environment this component belongs to.
 */
extern "C" void SAL_CALL component_getImplementationEnvironment(const sal_Char ** ppEnvTypeName, uno_Environment ** ppEnv)
{
	*ppEnvTypeName = CPPU_CURRENT_LANGUAGE_BINDING_NAME;
}

/**
 * This function creates an implementation section in the registry and another subkey
 *
 * for each supported service.
 * @param pServiceManager   the service manager
 * @param pRegistryKey      the registry key
 */
extern "C" sal_Bool SAL_CALL component_writeInfo(void * pServiceManager, void * pRegistryKey)
{
	sal_Bool result = sal_False;

	if (pRegistryKey)
	{
		try
		{
			Reference< XRegistryKey > xNewKey(
				reinterpret_cast< XRegistryKey * >( pRegistryKey )->createKey(
					OUString( RTL_CONSTASCII_USTRINGPARAM("/" IMPLNAME "/UNO/SERVICES") ) ) );
			
			const Sequence< OUString > & rSNL =
				MyGrammarChecker::getSupportedServiceNames_Static();
			const OUString * pArray = rSNL.getConstArray();
			for ( sal_Int32 nPos = rSNL.getLength(); nPos--; )
				xNewKey->createKey( pArray[nPos] );
			
			return sal_True;
		}
		catch (InvalidRegistryException &)
		{
			// we should not ignore exceptions
		}
	}
	return result;
}

extern "C" void * SAL_CALL component_getFactory(
    sal_Char const * implName, lang::XMultiServiceFactory * xMgr, void * )
{
	// printf("init esperantilo\n");
    Reference< lang::XSingleComponentFactory > xFactory;
    if (0 == ::rtl_str_compare( implName, IMPLNAME ))
    {
		Sequence < OUString > aRet(1);
    	OUString* pArray = aRet.getArray();
    	pArray[0] =  OUString ( RTL_CONSTASCII_USTRINGPARAM ( PROOFSERVICE ) );
        // create component factory for MyService1 implementation
        OUString serviceName( RTL_CONSTASCII_USTRINGPARAM(SERVICENAME) );
        xFactory = ::cppu::createSingleComponentFactory(
            create_KorektiloImpl,
            OUString( RTL_CONSTASCII_USTRINGPARAM(IMPLNAME) ),
            aRet );
    }
    if (xFactory.is())
        xFactory->acquire();
    return xFactory.get(); // return acquired interface pointer or null
}
