/***************************************************************************
 *   Copyright (C) 2008 by Artur Trzewik   *
 *   mail@xdobry.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>

#include <rtl/ustring.hxx>

#include <osl/diagnose.h>

#include <cppuhelper/bootstrap.hxx>
#include <iostream>

// generated c++ interfaces
#include <com/sun/star/lang/XComponent.hpp>
#include <com/sun/star/lang/XMultiComponentFactory.hpp>
#include <com/sun/star/registry/XImplementationRegistration.hpp>
#include <com/sun/star/linguistic2/XProofreader.hpp>

using namespace cppu;
using namespace com::sun::star::uno;
using namespace com::sun::star::lang;
using namespace com::sun::star::registry;
using namespace com::sun::star::linguistic2;

using namespace ::rtl;


//=======================================================================
int SAL_CALL main(int argc, char **argv)
{
	Reference< XSimpleRegistry > xReg = createSimpleRegistry(); 
	OSL_ENSURE( xReg.is(), "### cannot get service instance of \"com.sun.star.regiystry.SimpleRegistry\"!" );

	xReg->open(OUString::createFromAscii("esperantilo.uno.rdb"), sal_False, sal_False);
	OSL_ENSURE( xReg->isValid(), "### cannot open test registry \"esperantilo.uno.rdb\"!" );

	Reference< XComponentContext > xContext = bootstrap_InitialComponentContext(xReg);
	OSL_ENSURE( xContext.is(), "### cannot creage intial component context!" );

	Reference< XMultiComponentFactory > xMgr = xContext->getServiceManager();
	OSL_ENSURE( xMgr.is(), "### cannot get initial service manager!" );

	// register my counter component
	Reference< XImplementationRegistration > xImplReg(
		xMgr->createInstanceWithContext(OUString::createFromAscii("com.sun.star.registry.ImplementationRegistration"), xContext), UNO_QUERY);
	OSL_ENSURE( xImplReg.is(), "### cannot get service instance of \"com.sun.star.registry.ImplementationRegistration\"!" );

	std::cout<<"Here\n";

	if (xImplReg.is())
	{
		xImplReg->registerImplementation(
			OUString::createFromAscii("com.sun.star.loader.SharedLibrary"), // loader for component
#ifdef UNX
#ifdef MACOSX
			OUString::createFromAscii("esperantilo.uno.dylib"),		// component location
#else
			OUString::createFromAscii("esperantilo.uno.so"),		// component location
#endif
#else
			OUString::createFromAscii("esperantilo.uno.dll"),		// component location
#endif
			Reference< XSimpleRegistry >()	 // registry omitted,
											 // defaulting to service manager registry used
			);
	
		std::cout<<"Here2\n";
	
		// get a counter instance
		Reference< XInterface > xx ;
		xx = xMgr->createInstanceWithContext(OUString::createFromAscii("esperantilo.Korektilo"), xContext);
		Reference< XProofreader > korektilo( xx, UNO_QUERY );
		OSL_ENSURE( korektilo.is(), "### cannot get service instance of \"esperantilo.Korektilo\"!" );

		if (korektilo.is())
		{
			
			std::cout << "OK\n";
			if (korektilo->isSpellChecker()) {
				std::cout << "isSpellChecker yes\n";
			} else {
				std::cout << "isSpellChecker no\n";

			}	
		} else {
				std::cout << "wrong class\n";
		}
	}

	Reference< XComponent >::query( xContext )->dispose();
	return 0;
}
